@@ -176,16 +176,16 @@ static int auth_callback(const char *prompt, char *buf, size_t len,
   return 0;
 }
 
-static void log_callback(ssh_session session, int priority, const char *message,
+static void log_callback(int priority, const char *function, const char *buffer,
                          void *userdata)
 {
-  if (userdata == NULL) {
-    return;
-  }
+    if (userdata == NULL) {
+        return;
+    }
 
-  sftpProtocol *slave = (sftpProtocol *) userdata;
+    sftpProtocol *slave = (sftpProtocol *) userdata;
 
-  slave->log_callback(session, priority, message, userdata);
+    slave->log_callback(priority, function, buffer, userdata);
 }
 
 int sftpProtocol::auth_callback(const char *prompt, char *buf, size_t len,
@@ -235,13 +235,12 @@ int sftpProtocol::auth_callback(const char *prompt, char *buf, size_t len,
   return 0;
 }
 
-void sftpProtocol::log_callback(ssh_session session, int priority,
-                                const char *message, void *userdata)
+void sftpProtocol::log_callback(int priority, const char *function, const char *buffer,
+                                void *userdata)
 {
-  (void) session;
-  (void) userdata;
+    (void) userdata;
 
-  kDebug(KIO_SFTP_DB) << "[" << priority << "] " << message;
+    kDebug(KIO_SFTP_DB) << "[" << function << "] (" << priority << ") " << buffer;
 }
 
 int sftpProtocol::authenticateKeyboardInteractive(AuthInfo &info) {
@@ -464,11 +463,33 @@ sftpProtocol::sftpProtocol(const QByteArray &pool_socket, const QByteArray &app_
 
   mCallbacks->userdata = this;
   mCallbacks->auth_function = ::auth_callback;
-  if (getenv("KIO_SFTP_LOG_VERBOSITY")) {
-    mCallbacks->log_function = ::log_callback;
-  }
 
   ssh_callbacks_init(mCallbacks);
+
+  char *verbosity = getenv("KIO_SFTP_LOG_VERBOSITY");
+  if (verbosity != NULL) {
+    int level = atoi(verbosity);
+    int rc;
+
+    rc = ssh_set_log_level(level);
+    if (rc != SSH_OK) {
+      error(KIO::ERR_INTERNAL, i18n("Could not set log verbosity."));
+      return;
+    }
+
+    rc = ssh_set_log_userdata(this);
+    if (rc != SSH_OK) {
+      error(KIO::ERR_INTERNAL, i18n("Could not set log userdata."));
+      return;
+    }
+
+    rc = ssh_set_log_callback(::log_callback);
+    if (rc != SSH_OK) {
+      error(KIO::ERR_INTERNAL, i18n("Could not set log callback."));
+      return;
+    }
+  }
+
 }
 
 sftpProtocol::~sftpProtocol() {
@@ -503,7 +524,7 @@ bool sftpProtocol::sftpOpenConnection (const AuthInfo& info)
 {
   mSession = ssh_new();
   if (mSession == NULL) {
-    error(KIO::ERR_INTERNAL, i18n("Could not create a new SSH session."));
+    error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not create a new SSH session."));
     return false;
   }
 
@@ -514,39 +535,39 @@ bool sftpProtocol::sftpOpenConnection (const AuthInfo& info)
   // Set timeout
   int rc = ssh_options_set(mSession, SSH_OPTIONS_TIMEOUT, &timeout_sec);
   if (rc < 0) {
-    error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set a timeout."));
+    error(KIO::ERR_INTERNAL, i18n("Could not set a timeout."));
     return false;
   }
   rc = ssh_options_set(mSession, SSH_OPTIONS_TIMEOUT_USEC, &timeout_usec);
   if (rc < 0) {
-    error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set a timeout."));
+    error(KIO::ERR_INTERNAL, i18n("Could not set a timeout."));
     return false;
   }
 
   // Don't use any compression
   rc = ssh_options_set(mSession, SSH_OPTIONS_COMPRESSION_C_S, "none");
   if (rc < 0) {
-    error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set compression."));
+    error(KIO::ERR_INTERNAL, i18n("Could not set compression."));
     return false;
   }
 
   rc = ssh_options_set(mSession, SSH_OPTIONS_COMPRESSION_S_C, "none");
   if (rc < 0) {
-    error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set compression."));
+    error(KIO::ERR_INTERNAL, i18n("Could not set compression."));
     return false;
   }
 
   // Set host and port
   rc = ssh_options_set(mSession, SSH_OPTIONS_HOST, mHost.toUtf8().constData());
   if (rc < 0) {
-    error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set host."));
+    error(KIO::ERR_INTERNAL, i18n("Could not set host."));
     return false;
   }
 
   if (mPort > 0) {
     rc = ssh_options_set(mSession, SSH_OPTIONS_PORT, &mPort);
     if (rc < 0) {
-        error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set port."));
+        error(KIO::ERR_INTERNAL, i18n("Could not set port."));
       return false;
     }
   }
@@ -555,16 +576,7 @@ bool sftpProtocol::sftpOpenConnection (const AuthInfo& info)
   if (!info.username.isEmpty()) {
     rc = ssh_options_set(mSession, SSH_OPTIONS_USER, info.username.toUtf8().constData());
     if (rc < 0) {
-      error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set username."));
-      return false;
-    }
-  }
-
-  char* verbosity = getenv("KIO_SFTP_LOG_VERBOSITY");
-  if (verbosity) {
-    rc = ssh_options_set(mSession, SSH_OPTIONS_LOG_VERBOSITY_STR, verbosity);
-    if (rc < 0) {
-      error(KIO::ERR_OUT_OF_MEMORY, i18n("Could not set log verbosity."));
+      error(KIO::ERR_INTERNAL, i18n("Could not set username."));
       return false;
     }
   }
@@ -632,8 +644,10 @@ void sftpProtocol::openConnection() {
   QString msg;     // msg for dialog box
   QString caption; // dialog box caption
   unsigned char *hash = NULL; // the server hash
+  ssh_key srv_pubkey;
   char *hexa;
-  int rc, state, hlen;
+  size_t hlen;
+  int rc, state;
 
   // Attempt to start a ssh session and establish a connection with the server.
   if (!sftpOpenConnection(info)) {
@@ -643,13 +657,25 @@ void sftpProtocol::openConnection() {
   kDebug(KIO_SFTP_DB) << "Getting the SSH server hash";
 
   /* get the hash */
-  hlen = ssh_get_pubkey_hash(mSession, &hash);
-  if (hlen < 0) {
+  rc = ssh_get_publickey(mSession, &srv_pubkey);
+  if (rc < 0) {
     error(KIO::ERR_SLAVE_DEFINED, QString::fromUtf8(ssh_get_error(mSession)));
     closeConnection();
     return;
   }
 
+  rc = ssh_get_publickey_hash(srv_pubkey,
+                              SSH_PUBLICKEY_HASH_SHA1,
+                              &hash,
+                              &hlen);
+  ssh_key_free(srv_pubkey);
+  if (rc < 0) {
+    error(KIO::ERR_SLAVE_DEFINED,
+          i18n("Could not create hash from server public key"));
+    closeConnection();
+    return;
+  }
+
   kDebug(KIO_SFTP_DB) << "Checking if the SSH server is known";
 
   /* check the server public key hash */
@@ -658,7 +684,7 @@ void sftpProtocol::openConnection() {
     case SSH_SERVER_KNOWN_OK:
       break;
     case SSH_SERVER_FOUND_OTHER:
-      delete hash;
+      ssh_string_free_char((char *)hash);
       error(KIO::ERR_SLAVE_DEFINED, i18n("The host key for this server was "
             "not found, but another type of key exists.\n"
             "An attacker might change the default server key to confuse your "
@@ -668,7 +694,7 @@ void sftpProtocol::openConnection() {
       return;
     case SSH_SERVER_KNOWN_CHANGED:
       hexa = ssh_get_hexa(hash, hlen);
-      delete hash;
+      ssh_string_free_char((char *)hash);
       /* TODO print known_hosts file, port? */
       error(KIO::ERR_SLAVE_DEFINED, i18n("The host key for the server %1 has changed.\n"
           "This could either mean that DNS SPOOFING is happening or the IP "
@@ -676,18 +702,18 @@ void sftpProtocol::openConnection() {
           "The fingerprint for the key sent by the remote host is:\n %2\n"
           "Please contact your system administrator.\n%3",
           mHost, QString::fromUtf8(hexa), QString::fromUtf8(ssh_get_error(mSession))));
-      delete hexa;
+      ssh_string_free_char(hexa);
       closeConnection();
       return;
     case SSH_SERVER_FILE_NOT_FOUND:
     case SSH_SERVER_NOT_KNOWN:
       hexa = ssh_get_hexa(hash, hlen);
-      delete hash;
+      ssh_string_free_char((char *)hash);
       caption = i18n("Warning: Cannot verify host's identity.");
       msg = i18n("The authenticity of host %1 cannot be established.\n"
         "The key fingerprint is: %2\n"
         "Are you sure you want to continue connecting?", mHost, hexa);
-      delete hexa;
+      ssh_string_free_char(hexa);
 
       if (KMessageBox::Yes != messageBox(WarningYesNo, msg, caption)) {
         closeConnection();
@@ -704,7 +730,7 @@ void sftpProtocol::openConnection() {
       }
       break;
     case SSH_SERVER_ERROR:
-      delete hash;
+      ssh_string_free_char((char *)hash);
       error(KIO::ERR_SLAVE_DEFINED, QString::fromUtf8(ssh_get_error(mSession)));
       return;
   }
@@ -732,7 +758,7 @@ void sftpProtocol::openConnection() {
   if (rc != SSH_AUTH_SUCCESS && (method & SSH_AUTH_METHOD_PUBLICKEY)) {
     kDebug(KIO_SFTP_DB) << "Trying to authenticate with public key";
     for(;;) {
-      rc = ssh_userauth_autopubkey(mSession, NULL);
+      rc = ssh_userauth_publickey_auto(mSession, NULL, NULL);
       if (rc == SSH_AUTH_ERROR) {
         kDebug(KIO_SFTP_DB) << "Public key authentication failed:" <<
                 QString::fromUtf8(ssh_get_error(mSession));
@@ -747,6 +773,19 @@ void sftpProtocol::openConnection() {
     }
   }
 
+  // Try to authenticate with GSSAPI
+  if (rc != SSH_AUTH_SUCCESS && (method & SSH_AUTH_METHOD_GSSAPI_MIC)) {
+      kDebug(KIO_SFTP_DB) << "Trying to authenticate with GSSAPI";
+      rc = ssh_userauth_gssapi(mSession);
+      if (rc == SSH_AUTH_ERROR) {
+          kDebug(KIO_SFTP_DB) << "Public key authentication failed:" <<
+                 QString::fromUtf8(ssh_get_error(mSession));
+          closeConnection();
+          error(KIO::ERR_COULD_NOT_LOGIN, i18n("Authentication failed."));
+          return;
+      }
+  }
+
   // Try to authenticate with keyboard interactive
   if (rc != SSH_AUTH_SUCCESS && (method & SSH_AUTH_METHOD_INTERACTIVE)) {
     kDebug(KIO_SFTP_DB) << "Trying to authenticate with keyboard interactive";
@@ -1364,12 +1403,12 @@ sftpProtocol::StatusCode sftpProtocol::sftpPut(const KUrl& url, int permissions,
             initialMode = 0644;
           }
 
-          kDebug(KIO_SFTP_DB) << "Trying to open:" << dest << ", mode=" << QString::number(initialMode);
+          kDebug(KIO_SFTP_DB) << "Trying to open:" << QString(dest) << ", mode=" << QString::number(initialMode);
           file = sftp_open(mSftp, dest.constData(), O_CREAT | O_TRUNC | O_WRONLY, initialMode);
         } // flags & KIO::Resume
 
         if (file == NULL) {
-          kDebug(KIO_SFTP_DB) << "COULD NOT WRITE " << dest
+          kDebug(KIO_SFTP_DB) << "COULD NOT WRITE " << QString(dest)
                               << ", permissions=" << permissions
                               << ", error=" << ssh_get_error(mSession);
           if (sftp_get_error(mSftp) == SSH_FX_PERMISSION_DENIED) {
